; Scheme 9 from Empty Space, Function Library
; By Nils M Holm, 2010,2012
; Placed in the Public Domain
;
; (canvas-draw canvas integer-X integer-Y char)       ==>  unspecific
; (canvas-draw-string canvas int-X int-Y string)      ==>  unspecific
; (canvas-dump canvas)                                ==>  vector
; (canvas-plot canvas integer-X integer-Y char)       ==>  unspecific
; (canvas-plot-line canvas X Y DX DY char)            ==>  unspecific
; (make-canvas int-X int-Y int-W int-H)               ==>  canvas
;
; (load-from-library "char-canvas.scm")
;
; This is a set of routines for drawing characters and lines on
; a scaled, character-based (a.k.a. "ASCII Art") canvas.
;
; MAKE-CANVAS creates a char canvas with a physical size of
; x=INT-X times y=INT-Y characters. The virtual size of the
; canvas is INT-W (width) times INT-H (height) "pixels". "Real
; coordinates" relate to the physical size of the canvas.
; "Virtual coordinates" are translated to real coordinates by
; scaling. Both types of coordinates are specified in X/Y
; notation. The origin 0/0 is at the lower left corner of the
; canvas. The new canvas will be filled with blanks initially.
;
; CANVAS-DRAW draws character CHAR at position INTEGER-X/INTEGER-Y.
; It uses real coordinates. CANVAS-DRAWSTRING draws a string
; instead of a single character. When the X or Y coordinate is
; outside of the canvas, C will not be drawn. When STRING extends
; beyond the limits of the canvas, it will be clipped.
;
; CANVAS-PLOT draws the character CHAR at the virtual position
; INTEGER-X/INTEGER-Y. CANVAS-PLOT-LINE draws a line from the
; virtual position X/Y to DX/DY using the character CHAR. All
; arguments must be integers. Lines originating or extending
; outside of the canvas will be clipped.
;
; CANVAS-DUMP returns a vector of strings that contain the
; characters written to the canvas. The vector indexes are the
; Y-coordinates, the string offsets the X-coordinates.
;
; Example:   (let ((c (make-canvas 10 5 10 10)))
;              (canvas-plot-line c 0 9 9 0 #\#)
;              (canvas-plot-line c 0 0 9 9 #\*)
;              (canvas-dump c))                   ==>  #("##      **"
;                                                        "  ##  **  "
;                                                        "    **    "
;                                                        "  **  ##  "
;                                                        "**      ##")

(module (s9fes char-canvas)

(;export
  ;
  make-canvas
  canvas?
  canvas-width canvas-height
  canvas-virtual-width canvas-virtual-height
  ;
  canvas-dump
  canvas-clear
  ;
  canvas-draw
  canvas-draw-string
  ;
  canvas-plot
  canvas-plot-line
  canvas-plot-lines)

(import scheme utf8)
(import (chicken base))
(import (chicken fixnum))
(import (chicken type))

(define-type canvas (struct canvas))

(: make-canvas (fixnum fixnum #!optional integer integer -> canvas))

(: canvas-width (canvas --> fixnum))
(: canvas-height (canvas --> fixnum))
(: canvas-virtual-width (canvas --> integer))
(: canvas-virtual-height (canvas --> integer))

(: canvas-clear (canvas -> void))

(: canvas-dump (canvas -> (list-of string)))

(: canvas-draw (canvas fixnum fixnum char -> void))
(: canvas-draw-string (canvas fixnum fixnum string -> void))

(: canvas-plot (canvas integer integer char -> void))
(: canvas-plot-line (canvas integer integer integer integer char -> void))
(: canvas-plot-lines (canvas (list-of integer) char -> void))

;;

;srfi-133
(define (vector->string v) (list->string (vector->list v)))

;fx-utils
(define (fxnegative? n) (fx> 0 n))
(define (fxabs n) (if (fxnegative? n) (fxneg n) n))

;moremacros
(define-syntax swap!
  (syntax-rules ()
    ((swap! ?a ?b)
      (let ((_tmp ?a))
        (set! ?a ?b)
        (set! ?b _tmp)) ) ) )

;;

(define-record-type canvas
  (make-canvas* width height x-scale y-scale cmap)
  canvas?
  (width canvas-width)
  (height canvas-height)
  (x-scale canvas-x-scale)
  (y-scale canvas-y-scale)
  (cmap canvas-cmap set-canvas-cmap!))

(define (scale-x canvas x) (floor (* x (canvas-x-scale canvas))))
(define (scale-y canvas y) (floor (* y (canvas-y-scale canvas))))

(define (make-storage x-max y-max)
  (let ((v (make-vector y-max)))
    (do ((i 0 (fx+ i 1)))
        ((fx= i y-max))
      (vector-set! v i (make-vector x-max #\space)))
    v) )

;;

(define (make-canvas x-max y-max #!optional (v-x-max x-max) (v-y-max y-max))
  (make-canvas* x-max y-max
                (/ x-max v-x-max) (/ y-max v-y-max)
                (make-storage x-max y-max)) )

(define (canvas-virtual-width canvas)
  (* (canvas-width canvas) (/ (canvas-x-scale canvas))) )

(define (canvas-virtual-height canvas)
  (* (canvas-height canvas) (/ (canvas-y-scale canvas))) )

;do not expose the canvas
(define (canvas-dump canvas)
  (map vector->string (vector->list (canvas-cmap canvas))) )

(define (canvas-clear canvas)
  (set-canvas-cmap! canvas
                    (make-storage (canvas-width canvas) (canvas-height canvas))) )

;;

(define (canvas-draw canvas x y c)
  (let* ((cmap (canvas-cmap canvas))
         (k (canvas-height canvas)))
    (when (and (fx<= 0 y) (fx<= y (fx- k 1))
               (fx<= 0 x) (fx<= x (fx- (canvas-width canvas)
                                       1)))
      (vector-set! (vector-ref cmap (fx- (fx- k y) 1)) x c))) )

(define (canvas-draw-string canvas x y s)
  (let* ((ks (string-length s))
         (line (vector-ref (canvas-cmap canvas)
                           (fx- (fx- (canvas-height canvas)
                                     y)
                                1)))
         (kw (canvas-width canvas)) )
    (do ((x x (fx+ 1 x))
         (i 0 (fx+ 1 i)))
        ((or (fx>= i ks)
             (fx>= x kw)))
      (vector-set! line x (string-ref s i)))) )

;;

(define (canvas-plot canvas x y c)
  (canvas-draw canvas (scale-x canvas x) (scale-y canvas y) c) )

(define (canvas-plot-line canvas x0 y0 xn yn c)
  (let ((steep (> (abs (- yn y0))
                  (abs (- xn x0))))
        (x0 (scale-x canvas x0))
        (y0 (scale-y canvas y0))
        (xn (scale-x canvas xn))
        (yn (scale-y canvas yn)) )
    (when steep
      (swap! x0 y0)
      (swap! xn yn))
    (when (fx> x0 xn)
      (swap! x0 xn)
      (swap! y0 yn))
    (let ((dx (fx- xn x0))
          (dy (fxabs (fx- yn y0))))
      (let ((de (/ dy dx))
            (ys (if (fx< y0 yn) 1 -1)))
        (let plot ((x x0)
                   (y y0)
                   (e 0))
          (when (fx<= x xn)
            (if steep
                (canvas-draw canvas y x c)
                (canvas-draw canvas x y c))
            (let ((e (+ e de)))
              (when (<= 1/2 e)
                (set! y (fx+ y ys))
                (set! e (- e 1)))
              (plot (fx+ x 1)
                    y
                    e))))))) )

(define (canvas-plot-lines canvas ls c)
  (unless (null? ls)
    (assert (even? (length ls)) 'canvas-plot-lines "must be list-of x y ..." ls)
    (let ((x0 (car ls)) (y0 (cadr ls)))
      (if (fx= 2 (length ls))
          ;then single point
          (canvas-plot canvas x0 y0 c)
          ;else at least one line
          (let loop ((x0 x0) (y0 y0) (ls (cddr ls)))
            (let ((xn (car ls)) (yn (cadr ls)) (ls (cddr ls)))
              (canvas-plot-line canvas x0 y0 xn yn c)
              (unless (null? ls)
                (loop xn yn ls) ) ) ) ) ) ) )

) ;module (s9fes char-canvas)
